﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using devDept.Eyeshot;
using devDept.Eyeshot.Entities;
using devDept.Geometry;
using OpenGL;

namespace WindowsApplication1
{
    public class MyMesh : Mesh
    {
        private ViewportLayout vp;
        public MyMesh(ViewportLayout vp, meshNatureType meshNature)
            : base(meshNature)
        {
            this.vp = vp;
        }

        private int vertexBuffer, indexBuffer;
        private int indexBufferSelected;
        int verticesSize, normalsSize, indicesSize;

        public List<int> _selectedQuads = new List<int>();

        // Gets or sets the list of selected quads 
        public List<int> SelectedQuads
        {
            get { return _selectedQuads; }
            set
            {
                _selectedQuads = value;
                RegenMode = regenType.CompileOnly;
            }
        }
        public override void UpdateNormals()
        {
            // We keep the normals per Quad, don't let Eyeshot compute them 

            //base.UpdateNormals(); 
        }

        void CreateVBOForSelected()
        {
            // Creates the VertexBufferObjects of the selected quads 

            if (indexBufferSelected != 0)
                ClearBuffer(ref indexBufferSelected);

            if (SelectedQuads.Count == 0)
                return;

            // Create the buffer object 
            int[] buffer = gl.GenBuffersARB(1);

            indexBufferSelected = buffer[0];

            // Only the quads specified by the SelectedQuads array 
            int[] myIndices = null;

            myIndices = new int[SelectedQuads.Count * 4];

            int count = 0;
            for (int i = 0; i < SelectedQuads.Count; i++)
            {
                int indexTriangle = SelectedQuads[i] * 2;
                IndexTriangle tri = Triangles[indexTriangle];
                myIndices[count++] = tri.V1;
                myIndices[count++] = tri.V2;
                myIndices[count++] = tri.V3;
                myIndices[count++] = Triangles[indexTriangle + 1].V3;
            }

            indicesSize = myIndices.Length * sizeof(int);

            // Bind the element buffer object and copy the indices to it 
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, indexBufferSelected);
            gl.BufferData(gl.ELEMENT_ARRAY_BUFFER_ARB, indicesSize, myIndices, gl.STATIC_DRAW_ARB);

            // Unbind the buffer objects 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, 0);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, 0);
        }

        void CreateVBO()
        {
            // Use Vertex Buffer Objects for drawing the quads 
            int count;
            if (vertexBuffer != 0)
            {
                ClearBuffer(ref indexBuffer);
                ClearBuffer(ref vertexBuffer);
            }

            // Create the 2 buffer objects 
            int[] buffer = gl.GenBuffersARB(2);

            vertexBuffer = buffer[0];
            indexBuffer = buffer[1];

            // Fill the array with the indices of the quads vertices 

            int[] myIndices = null;

            myIndices = new int[2 * Triangles.Length];

            count = 0;
            for (int i = 0; i < Triangles.Length; i += 2)
            {
                IndexTriangle tri = Triangles[i];
                myIndices[count++] = tri.V1;
                myIndices[count++] = tri.V2;
                myIndices[count++] = tri.V3;
                myIndices[count++] = Triangles[i + 1].V3; // fourth vertex of the quad 
            }

            indicesSize = myIndices.Length * sizeof(int);

            // Bind the element buffer object and copy the indices to it 
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, indexBuffer);
            gl.BufferData(gl.ELEMENT_ARRAY_BUFFER_ARB, indicesSize, myIndices, gl.STATIC_DRAW_ARB);

            // The normals and colors array must have the same number of elements as the vertices array 
            verticesSize = Vertices.Length * 3 * sizeof(float);
            normalsSize = Vertices.Length * 3 * sizeof(float);

            // Bind the array buffer object for the vertices, normals and colors and allocate its size 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, vertexBuffer);
            gl.BufferData(gl.ARRAY_BUFFER_ARB, verticesSize + normalsSize, IntPtr.Zero,
                          gl.STREAM_DRAW_ARB);

            // Copy the data of the mesh to arrays of simple data types to pass to the buffer objects 
            float[] myVertices = new float[Vertices.Length * 3];
            float[] myNormals = new float[Vertices.Length * 3];

            count = 0;
            int quadIndex = 0;

            for (int i = 0; i < Vertices.Length; i++)
            {
                Point3D pt = (Point3D)Vertices[i];

                // Since there are 4 vertices per quad composing the quad, every 4 vertices use the same normal 
                if (i % 4 == 0 && i > 0)
                    quadIndex++;

                Vector3D n = Normals[quadIndex];

                myVertices[count] = (float)pt.X;
                myNormals[count] = (float)n.X;
                count++;

                myVertices[count] = (float)pt.Y;
                myNormals[count] = (float)n.Y;
                count++;

                myVertices[count] = (float)pt.Z;
                myNormals[count] = (float)n.Z;
                count++;
            }

            // Bind the array buffer object and copy the data to it 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, vertexBuffer);
            gl.BufferSubData(gl.ARRAY_BUFFER_ARB, 0, verticesSize, myVertices);
            gl.BufferSubData(gl.ARRAY_BUFFER_ARB, verticesSize, normalsSize, myNormals);

            // Unbind the buffer objects 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, 0);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, 0);
        }

        void DrawVBO(int myIndexBuffer, int nIndices)
        {
            // Draws the Vertex Buffer Object 
            // Bind the buffer objects 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, vertexBuffer);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, myIndexBuffer);

            // Activate the necessary arrays 
            gl.EnableClientState(gl.VERTEX_ARRAY);
            gl.EnableClientState(gl.NORMAL_ARRAY);
            
            // Specify the pointers to vertex data (the last parameter is the offset in the buffer array) 
            gl.VertexPointer(3, gl.FLOAT, 0, 0);
            gl.NormalPointer(gl.FLOAT, 0, verticesSize);

            // Draw the triangles 
            gl.DrawElements(gl.QUADS, nIndices, gl.UNSIGNED_INT, 0);

            // Disable the arrays 
            gl.DisableClientState(gl.VERTEX_ARRAY);
            gl.DisableClientState(gl.NORMAL_ARRAY);
            
            // Unbind the buffer objects 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, 0);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, 0);
        }

        protected override void DrawForShadow()
        {
            DrawSelected(null);
        }

        protected override void DrawForSelection(DrawParams data)
        {
            // Draws the quads with the color-coding needed for visibility computation 
            gl.Begin(gl.QUADS);
            int quadIndex = 0;
            for (int i = 0; i < Triangles.Length; i += 2)
            {
                vp.SetColorDrawForSelection(quadIndex++);

                IndexTriangle tri = Triangles[i];
                Vertices[tri.V1].DrawGL();
                Vertices[tri.V2].DrawGL();
                Vertices[tri.V3].DrawGL();
                Vertices[Triangles[i + 1].V3].DrawGL();
            }

            gl.End();
        }

        protected override void DrawSelected(DrawParams drawParams)
        {
            // Draws the Quads as simple as possible, without normals or the selected ones 

            if (vboSupported)
                DrawBasicVBO(indexBuffer, 2 * Triangles.Length);
            else
                base.Draw(drawParams);
        }

        private void DrawSelectedQuads(bool wireframe)
        {
            // Draws the selected Quads over the other quads 

            if (SelectedQuads.Count == 0)
                return;

            gl.DepthFunc(gl.EQUAL);
            gl.PushAttrib(gl.CURRENT_BIT);

            if (wireframe)
                ViewportLayout.SetColorWireframe(System.Drawing.Color.Yellow);
            else
                ViewportLayout.SetColorShaded(entityNatureType.Polygon, System.Drawing.Color.Yellow, true);

            if (vboSupported)
                DrawVBO(indexBufferSelected, SelectedQuads.Count * 4);
            else
                gl.CallList(drawSelectedList);

            gl.PopAttrib();
            gl.DepthFunc(gl.LESS);
        }

        void DrawBasicVBO(int myIndexBuffer, int nIndices)
        {
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, vertexBuffer);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, myIndexBuffer);

            // Activate the necessary arrays 
            gl.EnableClientState(gl.VERTEX_ARRAY);

            // Specify the pointers to vertex data (the last parameter is the offset in the buffer array) 
            gl.VertexPointer(3, gl.FLOAT, 0, 0);

            // Draw the triangles 
            gl.DrawElements(gl.QUADS, nIndices, gl.UNSIGNED_INT, 0);

            // Disable the arrays 
            gl.DisableClientState(gl.VERTEX_ARRAY);
            
            // Unbind the buffer objects 
            gl.BindBufferARB(gl.ARRAY_BUFFER_ARB, 0);
            gl.BindBufferARB(gl.ELEMENT_ARRAY_BUFFER_ARB, 0);
        }

        public override void Dispose()
        {
            // Deletes the buffer objects 
            ClearVBO();
            base.Dispose();
        }

        private void ClearVBO()
        {
            ClearBuffer(ref indexBufferSelected);
            ClearBuffer(ref indexBuffer);
            ClearBuffer(ref vertexBuffer);
        }

        private static void ClearBuffer(ref int buffer)
        {
            gl.DeleteBuffersARB(new int[] { buffer });

            buffer = 0;
        }

        private bool firstTime = true;
        private bool vboSupported;

        public override void Compile(CompileParams data)
        {
            if (firstTime && ViewportLayout.OpenglExtensions.Contains("ARB_vertex_buffer_object"))
            {
                firstTime = false;
                vboSupported = true;
            }

            CompileEdges(0);

            if (vboSupported)

                CreateVBO();

            else
            {
                ViewportLayout.GenListName(ref drawList);
                gl.NewList(drawList, gl.COMPILE);

                gl.Begin(gl.QUADS);

                for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)

                    DrawTriangle(countQuad, Triangles[i], Triangles[i + 1]);

                gl.End();

                gl.EndList();
            }

            CompileSelectedQuads();
        }

        private void CompileSelectedQuads()
        {
            // Compiles the selected quads 
            if (vboSupported)

                CreateVBOForSelected();

            else
            {
                ViewportLayout.GenListName(ref drawSelectedList);
                gl.NewList(drawSelectedList, gl.COMPILE);

                gl.Begin(gl.QUADS);

                // Draw selected triangles 
                for (int i = 0; i < SelectedQuads.Count; i++)
                {
                    int triangleIndex = SelectedQuads[i] * 2;
                    DrawTriangle(SelectedQuads[i], Triangles[triangleIndex], Triangles[triangleIndex + 1]);
                }

                gl.End();
                gl.EndList();
            }
        }

        private void DrawTriangle(int normalIndex, IndexTriangle tri, IndexTriangle nextTri)
        {
            Normals[normalIndex].DrawGL();
            Vertices[tri.V1].DrawGL();
            Vertices[tri.V2].DrawGL();
            Vertices[tri.V3].DrawGL();
            Vertices[nextTri.V3].DrawGL();
        }

        protected override void Draw(DrawParams data)
        {
            if (vboSupported)
                DrawVBO(indexBuffer, 2 * Triangles.Length);
            else
                base.Draw(data);

            DrawSelectedQuads(false);
        }

        protected override void Render(RenderParams data)
        {
            Draw(data);
        }

        protected override void DrawWireframe(DrawParams drawParams)
        {
            if (vboSupported)
                DrawVBO(indexBuffer, 2 * Triangles.Length);
            else
                base.DrawWireframe(drawParams);

            DrawSelectedQuads(true);
        }

        protected override void DrawIsocurves()
        {
            //base.DrawIsocurves(); 
        }

        protected override bool IsCrossing(FrustumParams data)
        {
            SelectedQuads = new List<int>();

            bool res = base.IsCrossing(data);

            CompileSelectedQuads();

            return res;
        }

        protected override bool InsideOrCrossingFrustum(FrustumParams data)
        {
            // Computes the quads that are inside or crossing the selection planes

            Point3D[] myVertices = new Point3D[4];

            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];
                myVertices[0] = Vertices[tri.V1];
                myVertices[1] = Vertices[tri.V2];
                myVertices[2] = Vertices[tri.V3];
                myVertices[3] = Vertices[tri2.V3];

                if (Utility.InsideOrCrossingFrustumQuad(myVertices[0], myVertices[1], myVertices[2], myVertices[3], data.Frustum))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }

            return false;

        }

        protected override bool ThroughTriangle(FrustumParams data)
        {
            SelectedQuads.Sort();

            Point3D[] myVertices = new Point3D[4];
            // Since it's usually called 
            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                if (SelectedQuads.BinarySearch(countQuad) >= 0)
                    continue;

                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];
                myVertices[0] = Vertices[tri.V1];
                myVertices[1] = Vertices[tri.V2];
                myVertices[2] = Vertices[tri.V3];
                myVertices[3] = Vertices[tri2.V3];

                if (ThroughTriangleQuad(data, myVertices))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }
            return false;
        }

        protected override bool IsCrossingScreenPolygon(ScreenPolygonParams data)
        {
            SelectedQuads = new List<int>();

            bool res = base.IsCrossingScreenPolygon(data);

            CompileSelectedQuads();

            return res;
        }


        protected override bool InsideOrCrossingScreenPolygon(ScreenPolygonParams data)
        {
            Point3D[] myVertices = new Point3D[4];

            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];
                myVertices[0] = Vertices[tri.V1];
                myVertices[1] = Vertices[tri.V2];
                myVertices[2] = Vertices[tri.V3];
                myVertices[3] = Vertices[tri2.V3];

                if (UtilityEx.InsideOrCrossingScreenPolygonQuad(myVertices[0], myVertices[1], myVertices[2], myVertices[3], data))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }

            return false;
        }

        protected override bool ThroughTriangleScreenPolygon(ScreenPolygonParams data)
        {
            SelectedQuads.Sort();

            Point3D[] myVertices = new Point3D[4];
            // Since it's usually called 
            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                if (SelectedQuads.BinarySearch(countQuad) >= 0)
                    continue;

                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];
                myVertices[0] = Vertices[tri.V1];
                myVertices[1] = Vertices[tri.V2];
                myVertices[2] = Vertices[tri.V3];
                myVertices[3] = Vertices[tri2.V3];

                if (ThroughTriangleScreenPolygonQuad(myVertices, data))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }
            return false;
        }

        protected override bool AllVerticesInFrustum(FrustumParams data)
        {
            SelectedQuads = new List<int>();

            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];

                if (Camera.IsInFrustum(Vertices[tri.V1], data.Frustum) && Camera.IsInFrustum(Vertices[tri.V2], data.Frustum) &&
                    Camera.IsInFrustum(Vertices[tri.V3], data.Frustum) && Camera.IsInFrustum(Vertices[tri2.V3], data.Frustum))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }

            CompileSelectedQuads();

            return false;
        }

        protected override bool AllVerticesInScreenPolygon(ScreenPolygonParams data)
        {
            SelectedQuads = new List<int>();

            Point3D[] verts = new Point3D[4];
            for (int i = 0, countQuad = 0; i < Triangles.Length; i += 2, countQuad++)
            {
                IndexTriangle tri = Triangles[i];
                IndexTriangle tri2 = Triangles[i + 1];

                verts[0] = Vertices[tri.V1];
                verts[1] = Vertices[tri.V2];
                verts[2] = Vertices[tri.V3];
                verts[3] = Vertices[tri2.V1];

                if (UtilityEx.AllVerticesInScreenPolygon(data, verts, 4))
                {
                    // Since I'm simulating the use of quads, I select also the other triangle of the quad
                    SelectedQuads.Add(countQuad);
                }
            }

            CompileSelectedQuads();

            return false;
        }

        internal void SelectQuads(int[] indices)
        {
            SelectedQuads = new List<int>(indices);
            CompileSelectedQuads();
        }
    }
}
